@@ -10,11 +10,15 @@
 #include "loggingcategory.h"
 #include <QAbstractItemModel>
 #include <QGuiApplication>
+#include <QJsonDocument>
+#include <QJsonObject>
+#include <QJsonParseError>
 #include <QPropertyAnimation>
 #include <QQmlComponent>
 #include <QQmlContext>
 #include <QQmlEngine>
 #include <QStyleHints>
+#include <algorithm>
 
 #include "platform/units.h"
 
@@ -58,64 +62,14 @@ QmlComponentsPool *QmlComponentsPoolSingleton::instance(QQmlEngine *engine)
 
 QmlComponentsPool::QmlComponentsPool(QQmlEngine *engine)
     : QObject(engine)
+    , m_separatorComponent(engine)
 {
-    QQmlComponent component(engine);
+    m_separatorComponent.loadFromModule("org.kde.kirigami.layouts.private", "ColumnViewSeparator");
 
-    /* clang-format off */
-    component.setData(QByteArrayLiteral(R"(
-import QtQuick
-import org.kde.kirigami as Kirigami
-
-QtObject {
-    readonly property Component leadingSeparator: Kirigami.Separator {
-        property Item column
-        property bool inToolBar
-        property Kirigami.ColumnView view
-
-        // positioning trick to hide the very first separator
-        visible: {
-            if (!view || !view.separatorVisible) {
-                return false;
-            }
-
-            return view && (LayoutMirroring.enabled
-                ? view.contentX + view.width > column.x + column.width
-                : view.contentX < column.x);
-        }
-
-        anchors.top: column.top
-        anchors.left: column.left
-        anchors.bottom: column.bottom
-        anchors.topMargin: inToolBar ? Kirigami.Units.largeSpacing : 0
-        anchors.bottomMargin: inToolBar ? Kirigami.Units.largeSpacing : 0
-        Kirigami.Theme.colorSet: Kirigami.Theme.Header
-        Kirigami.Theme.inherit: false
+    if (m_separatorComponent.isError()) {
+        qCFatal(KirigamiLayoutsLog) << m_separatorComponent.errors();
     }
 
-    readonly property Component trailingSeparator: Kirigami.Separator {
-        property Item column
-
-        anchors.top: column.top
-        anchors.right: column.right
-        anchors.bottom: column.bottom
-        Kirigami.Theme.colorSet: Kirigami.Theme.Header
-        Kirigami.Theme.inherit: false
-    }
-}
-)"), QUrl(QStringLiteral("columnview.cpp")));
-    /* clang-format on */
-
-    m_instance = component.create();
-    // qCWarning(KirigamiLayoutsLog)<<component.errors();
-    Q_ASSERT(m_instance);
-    m_instance->setParent(this);
-
-    m_leadingSeparatorComponent = m_instance->property("leadingSeparator").value<QQmlComponent *>();
-    Q_ASSERT(m_leadingSeparatorComponent);
-
-    m_trailingSeparatorComponent = m_instance->property("trailingSeparator").value<QQmlComponent *>();
-    Q_ASSERT(m_trailingSeparatorComponent);
-
     m_units = engine->singletonInstance<Kirigami::Platform::Units *>("org.kde.kirigami.platform", "Units");
     Q_ASSERT(m_units);
 
@@ -187,13 +141,38 @@ bool ColumnViewAttached::fillWidth() const
 
 qreal ColumnViewAttached::reservedSpace() const
 {
-    return m_reservedSpace;
+    if (m_customReservedSpace) {
+        return m_reservedSpace;
+    } else if (!m_view) {
+        return 0.0;
+    } else if (m_view->columnResizeMode() == ColumnView::FixedColumns) {
+        return m_view->columnWidth();
+    } else {
+        QQuickItem *prevItem = m_view->get(m_index - 1);
+        QQuickItem *nextItem = m_view->get(m_index + 1);
+
+        if (prevItem) {
+            ColumnViewAttached *prevAttached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(prevItem, true));
+
+            if (!prevAttached->fillWidth()) {
+                return prevItem->width();
+            }
+        }
+        if (nextItem) {
+            ColumnViewAttached *prevAttached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(nextItem, true));
+
+            if (!prevAttached->fillWidth()) {
+                return nextItem->width();
+            }
+        }
+    }
+    return m_view->columnWidth();
 }
 
 void ColumnViewAttached::setReservedSpace(qreal space)
 {
     if (m_view) {
-        disconnect(m_view.data(), &ColumnView::columnWidthChanged, this, nullptr);
+        disconnect(m_view.data(), &ColumnView::columnWidthChanged, this, &ColumnViewAttached::reservedSpaceChanged);
     }
     m_customReservedSpace = true;
 
@@ -209,6 +188,71 @@ void ColumnViewAttached::setReservedSpace(qreal space)
     }
 }
 
+qreal ColumnViewAttached::minimumWidth() const
+{
+    return m_minimumWidth;
+}
+
+void ColumnViewAttached::setMinimumWidth(qreal width)
+{
+    if (qFuzzyCompare(width, m_minimumWidth)) {
+        return;
+    }
+
+    m_minimumWidth = width;
+
+    Q_EMIT minimumWidthChanged();
+
+    if (m_view) {
+        m_view->polish();
+    }
+}
+
+qreal ColumnViewAttached::preferredWidth() const
+{
+    qreal preferred = m_preferredWidth;
+    // Fallback to the item implicitWidth
+    if (preferred < 0) {
+        preferred = static_cast<QQuickItem *>(parent())->implicitWidth();
+    }
+    return std::max(m_minimumWidth, std::min(m_maximumWidth, preferred));
+}
+
+void ColumnViewAttached::setPreferredWidth(qreal width)
+{
+    if (qFuzzyCompare(width, m_preferredWidth)) {
+        return;
+    }
+
+    m_preferredWidth = width;
+
+    Q_EMIT preferredWidthChanged();
+
+    if (m_view) {
+        m_view->polish();
+    }
+}
+
+qreal ColumnViewAttached::maximumWidth() const
+{
+    return m_maximumWidth;
+}
+
+void ColumnViewAttached::setMaximumWidth(qreal width)
+{
+    if (qFuzzyCompare(width, m_maximumWidth)) {
+        return;
+    }
+
+    m_maximumWidth = width;
+
+    Q_EMIT maximumWidthChanged();
+
+    if (m_view) {
+        m_view->polish();
+    }
+}
+
 ColumnView *ColumnViewAttached::view()
 {
     return m_view;
@@ -233,11 +277,7 @@ void ColumnViewAttached::setView(ColumnView *view)
         });
     }
     if (!m_customReservedSpace && m_view) {
-        m_reservedSpace = m_view->columnWidth();
-        connect(m_view.data(), &ColumnView::columnWidthChanged, this, [this]() {
-            m_reservedSpace = m_view->columnWidth();
-            Q_EMIT reservedSpaceChanged();
-        });
+        connect(m_view.data(), &ColumnView::columnWidthChanged, this, &ColumnViewAttached::reservedSpaceChanged);
     }
 
     Q_EMIT viewChanged();
@@ -314,6 +354,41 @@ void ColumnViewAttached::setInViewport(bool inViewport)
     Q_EMIT inViewportChanged();
 }
 
+bool ColumnViewAttached::interactiveResizeEnabled() const
+{
+    return m_interactiveResizeEnabled;
+}
+
+void ColumnViewAttached::setInteractiveResizeEnabled(bool interactive)
+{
+    if (interactive == m_interactiveResizeEnabled) {
+        return;
+    }
+
+    m_interactiveResizeEnabled = interactive;
+
+    Q_EMIT interactiveResizeEnabledChanged();
+}
+
+bool ColumnViewAttached::interactiveResizing() const
+{
+    return m_interactiveResizing;
+}
+
+void ColumnViewAttached::setInteractiveResizing(bool interactive)
+{
+    if (interactive == m_interactiveResizing) {
+        return;
+    }
+
+    m_interactiveResizing = interactive;
+
+    Q_EMIT interactiveResizingChanged();
+    if (!m_interactiveResizing && m_view) {
+        Q_EMIT m_view->savedStateChanged();
+    }
+}
+
 QQuickItem *ColumnViewAttached::globalHeader() const
 {
     return m_globalHeader;
@@ -487,12 +562,20 @@ qreal ContentItem::childWidth(QQuickItem *child)
         // DynamicColumns
     } else {
         // TODO:look for Layout size hints
-        qreal width = child->implicitWidth();
+        qreal width = attached->preferredWidth();
 
         if (width < 1.0) {
             width = m_columnWidth;
         }
 
+        if (attached->minimumWidth() >= 0 && attached->maximumWidth() >= 0) {
+            width = std::clamp(width, attached->minimumWidth(), attached->maximumWidth());
+        } else if (attached->minimumWidth() >= 0) {
+            width = std::max(width, attached->minimumWidth());
+        } else if (attached->maximumWidth() >= 0) {
+            width = std::min(width, attached->maximumWidth());
+        }
+
         return qRound(qMin(m_view->width(), width));
     }
 }
@@ -510,13 +593,26 @@ void ContentItem::layoutItems()
     m_rightPinnedSpace = 0;
 
     bool reverse = qApp->layoutDirection() == Qt::RightToLeft;
-    auto it = !reverse ? m_items.begin() : m_items.end();
+    auto it = !reverse ? m_items.begin() : m_items.end() - 1;
     int increment = reverse ? -1 : +1;
-    auto lastPos = reverse ? m_items.begin() : m_items.end();
+    auto lastPos = reverse ? m_items.begin() - 1 : m_items.end();
+
+    QQuickItem *previousChild = nullptr;
 
     for (; it != lastPos; it += increment) {
         // for (QQuickItem *child : std::as_const(m_items)) {
-        QQuickItem *child = reverse ? *(it - 1) : *it;
+        QQuickItem *child = *it;
+        QQuickItem *nextChild = nullptr;
+        if (reverse) {
+            if (it != m_items.begin()) {
+                nextChild = *(it - 1);
+            }
+        } else {
+            if ((it + 1) != m_items.end()) {
+                nextChild = *(it + 1);
+            }
+        }
+
         ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(child, true));
         if (child == m_globalHeaderParent || child == m_globalFooterParent) {
             continue;
@@ -527,7 +623,7 @@ void ContentItem::layoutItems()
                 QQuickItem *sep = nullptr;
                 int sepWidth = 0;
                 if (m_view->separatorVisible()) {
-                    sep = ensureTrailingSeparator(child);
+                    sep = ensureSeparator(previousChild, child, nextChild);
                     sepWidth = (sep ? sep->width() : 0);
                 }
                 const qreal width = childWidth(child);
@@ -540,20 +636,12 @@ void ContentItem::layoutItems()
                     header->setWidth(width + sepWidth);
                     header->setPosition(QPointF(pageX, .0));
                     header->setZ(2);
-                    if (m_view->separatorVisible()) {
-                        QQuickItem *sep = ensureTrailingSeparator(header);
-                        sep->setProperty("inToolBar", true);
-                    }
                 }
                 if (QQuickItem *footer = attached->globalFooter()) {
                     footerHeight = footer->isVisible() ? footer->height() : .0;
                     footer->setWidth(width + sepWidth);
                     footer->setPosition(QPointF(pageX, height() - footerHeight));
                     footer->setZ(2);
-                    if (m_view->separatorVisible()) {
-                        QQuickItem *sep = ensureTrailingSeparator(footer);
-                        sep->setProperty("inToolBar", true);
-                    }
                 }
 
                 child->setSize(QSizeF(width + sepWidth, height() - headerHeight - footerHeight));
@@ -572,44 +660,34 @@ void ContentItem::layoutItems()
                 const qreal width = childWidth(child);
                 qreal headerHeight = .0;
                 qreal footerHeight = .0;
+                if (m_view->separatorVisible()) {
+                    ensureSeparator(previousChild, child, nextChild);
+                }
                 if (QQuickItem *header = attached->globalHeader(); header && qmlEngine(header)) {
-                    if (m_view->separatorVisible()) {
-                        QQuickItem *sep = ensureLeadingSeparator(header);
-                        sep->setProperty("inToolBar", true);
-                    }
                     headerHeight = header->isVisible() ? header->height() : .0;
                     header->setWidth(width);
                     header->setPosition(QPointF(partialWidth, .0));
                     header->setZ(1);
-                    auto it = m_trailingSeparators.find(header);
-                    if (it != m_trailingSeparators.end()) {
+                    auto it = m_separators.find(header);
+                    if (it != m_separators.end()) {
                         it.value()->deleteLater();
-                        m_trailingSeparators.erase(it);
+                        m_separators.erase(it);
                     }
                 }
                 if (QQuickItem *footer = attached->globalFooter(); footer && qmlEngine(footer)) {
-                    if (m_view->separatorVisible()) {
-                        QQuickItem *sep = ensureLeadingSeparator(footer);
-                        sep->setProperty("inToolBar", true);
-                    }
                     footerHeight = footer->isVisible() ? footer->height() : .0;
                     footer->setWidth(width);
                     footer->setPosition(QPointF(partialWidth, height() - footerHeight));
                     footer->setZ(1);
-                    auto it = m_trailingSeparators.find(footer);
-                    if (it != m_trailingSeparators.end()) {
+                    // TODO: remove
+                    auto it = m_separators.find(footer);
+                    if (it != m_separators.end()) {
                         it.value()->deleteLater();
-                        m_trailingSeparators.erase(it);
+                        m_separators.erase(it);
                     }
                 }
 
                 child->setSize(QSizeF(width, height() - headerHeight - footerHeight));
-
-                auto it = m_trailingSeparators.find(child);
-                if (it != m_trailingSeparators.end()) {
-                    it.value()->deleteLater();
-                    m_trailingSeparators.erase(it);
-                }
                 child->setPosition(QPointF(partialWidth, headerHeight));
                 child->setZ(0);
 
@@ -623,9 +701,11 @@ void ContentItem::layoutItems()
             attached->setIndex(i++);
         }
 
-        implicitWidth += child->implicitWidth();
+        implicitWidth += attached->preferredWidth();
 
         implicitHeight = qMax(implicitHeight, child->implicitHeight());
+
+        previousChild = child;
     }
 
     setWidth(partialWidth);
@@ -656,7 +736,13 @@ void ContentItem::layoutPinnedItems()
     m_leftPinnedSpace = 0;
     m_rightPinnedSpace = 0;
 
-    for (QQuickItem *child : std::as_const(m_items)) {
+    QQuickItem *previousChild = nullptr;
+
+    for (auto it = m_items.constBegin(); it != m_items.constEnd(); it++) {
+        // for (QQuickItem *child : std::as_const(m_items)) {
+        QQuickItem *child = *it;
+        QQuickItem *nextChild = *(it + 1);
+        //  for (QQuickItem *child : std::as_const(m_items)) {
         ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(child, true));
 
         if (child->isVisible()) {
@@ -664,7 +750,7 @@ void ContentItem::layoutPinnedItems()
                 QQuickItem *sep = nullptr;
                 int sepWidth = 0;
                 if (m_view->separatorVisible()) {
-                    sep = ensureTrailingSeparator(child);
+                    sep = ensureSeparator(previousChild, child, nextChild);
                     sepWidth = (sep ? sep->width() : 0);
                 }
 
@@ -674,18 +760,10 @@ void ContentItem::layoutPinnedItems()
                 if (QQuickItem *header = attached->globalHeader()) {
                     headerHeight = header->isVisible() ? header->height() : .0;
                     header->setPosition(QPointF(pageX, .0));
-                    if (m_view->separatorVisible()) {
-                        QQuickItem *sep = ensureTrailingSeparator(header);
-                        sep->setProperty("inToolBar", true);
-                    }
                 }
                 if (QQuickItem *footer = attached->globalFooter()) {
                     footerHeight = footer->isVisible() ? footer->height() : .0;
                     footer->setPosition(QPointF(pageX, height() - footerHeight));
-                    if (m_view->separatorVisible()) {
-                        QQuickItem *sep = ensureTrailingSeparator(footer);
-                        sep->setProperty("inToolBar", true);
-                    }
                 }
                 child->setPosition(QPointF(pageX, headerHeight));
 
@@ -715,9 +793,21 @@ void ContentItem::updateVisibleItems()
             });
             attached->setInViewport(true);
             item->setEnabled(true);
+            if (attached->globalHeader()) {
+                attached->globalHeader()->setEnabled(true);
+            }
+            if (attached->globalFooter()) {
+                attached->globalFooter()->setEnabled(true);
+            }
         } else {
             attached->setInViewport(false);
             item->setEnabled(false);
+            if (attached->globalHeader()) {
+                attached->globalHeader()->setEnabled(false);
+            }
+            if (attached->globalFooter()) {
+                attached->globalFooter()->setEnabled(false);
+            }
         }
     }
 
@@ -754,11 +844,7 @@ void ContentItem::forgetItem(QQuickItem *item)
     disconnect(item, nullptr, this, nullptr);
     disconnect(item, nullptr, m_view, nullptr);
 
-    QQuickItem *separatorItem = m_leadingSeparators.take(item);
-    if (separatorItem) {
-        separatorItem->deleteLater();
-    }
-    separatorItem = m_trailingSeparators.take(item);
+    QQuickItem *separatorItem = m_separators.take(item);
     if (separatorItem) {
         separatorItem->deleteLater();
     }
@@ -766,11 +852,7 @@ void ContentItem::forgetItem(QQuickItem *item)
     if (QQuickItem *header = attached->globalHeader()) {
         header->setVisible(false);
         header->setParentItem(item);
-        separatorItem = m_leadingSeparators.take(header);
-        if (separatorItem) {
-            separatorItem->deleteLater();
-        }
-        separatorItem = m_trailingSeparators.take(header);
+        separatorItem = m_separators.take(header);
         if (separatorItem) {
             separatorItem->deleteLater();
         }
@@ -778,11 +860,7 @@ void ContentItem::forgetItem(QQuickItem *item)
     if (QQuickItem *footer = attached->globalFooter()) {
         footer->setVisible(false);
         footer->setParentItem(item);
-        separatorItem = m_leadingSeparators.take(footer);
-        if (separatorItem) {
-            separatorItem->deleteLater();
-        }
-        separatorItem = m_trailingSeparators.take(footer);
+        separatorItem = m_separators.take(footer);
         if (separatorItem) {
             separatorItem->deleteLater();
         }
@@ -802,42 +880,29 @@ void ContentItem::forgetItem(QQuickItem *item)
     Q_EMIT m_view->countChanged();
 }
 
-QQuickItem *ContentItem::ensureLeadingSeparator(QQuickItem *item)
+QQuickItem *ContentItem::ensureSeparator(QQuickItem *previousColumn, QQuickItem *column, QQuickItem *nextColumn)
 {
-    QQuickItem *separatorItem = m_leadingSeparators.value(item);
+    QQuickItem *separatorItem = m_separators.value(column);
 
     if (!separatorItem) {
         separatorItem = qobject_cast<QQuickItem *>(
-            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_leadingSeparatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
+            QmlComponentsPoolSingleton::instance(qmlEngine(column))->m_separatorComponent.beginCreate(QQmlEngine::contextForObject(column)));
         if (separatorItem) {
+            // Do NOT parent the separator to the column
+            // we are managing the lifetime of the separator on this side,
+            // therefore if is column itself deleting it, we will have a double dellete
             separatorItem->setParent(this);
-            separatorItem->setParentItem(item);
+            separatorItem->setParentItem(column);
             separatorItem->setZ(9999);
-            separatorItem->setProperty("column", QVariant::fromValue(item));
-            separatorItem->setProperty("view", QVariant::fromValue(m_view));
-            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_leadingSeparatorComponent->completeCreate();
-            m_leadingSeparators[item] = separatorItem;
+            separatorItem->setProperty("column", QVariant::fromValue(column));
+            QmlComponentsPoolSingleton::instance(qmlEngine(column))->m_separatorComponent.completeCreate();
+            m_separators[column] = separatorItem;
         }
     }
 
-    return separatorItem;
-}
-
-QQuickItem *ContentItem::ensureTrailingSeparator(QQuickItem *item)
-{
-    QQuickItem *separatorItem = m_trailingSeparators.value(item);
-
-    if (!separatorItem) {
-        separatorItem = qobject_cast<QQuickItem *>(
-            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_trailingSeparatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
-        if (separatorItem) {
-            separatorItem->setParent(this);
-            separatorItem->setParentItem(item);
-            separatorItem->setZ(9999);
-            separatorItem->setProperty("column", QVariant::fromValue(item));
-            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_trailingSeparatorComponent->completeCreate();
-            m_trailingSeparators[item] = separatorItem;
-        }
+    if (separatorItem) {
+        separatorItem->setProperty("previousColumn", QVariant::fromValue(previousColumn));
+        separatorItem->setProperty("nextColumn", QVariant::fromValue(nextColumn));
     }
 
     return separatorItem;
@@ -871,10 +936,6 @@ void ContentItem::itemChange(QQuickItem::ItemChange change, const QQuickItem::It
             });
         }
 
-        if (m_view->separatorVisible()) {
-            ensureLeadingSeparator(value.item);
-        }
-
         m_shouldAnimate = true;
         m_view->polish();
         Q_EMIT m_view->countChanged();
@@ -976,8 +1037,9 @@ ColumnView::ColumnView(QQuickItem *parent)
     m_contentItem = new ContentItem(this);
     // Prevent interactions outside of ColumnView bounds, and let it act as a viewport.
     setClip(true);
-    setAcceptedMouseButtons(Qt::LeftButton | Qt::BackButton | Qt::ForwardButton);
-    setAcceptTouchEvents(false); // Relies on synthetized mouse events
+    setAcceptedMouseButtons(Qt::BackButton | Qt::ForwardButton);
+    setAcceptHoverEvents(false);
+    setAcceptTouchEvents(true);
     setFiltersChildMouseEvents(true);
 
     connect(m_contentItem->m_slideAnim, &QPropertyAnimation::finished, this, [this]() {
@@ -1077,7 +1139,7 @@ void ColumnView::setCurrentIndex(int index)
                             width() - m_contentItem->m_rightPinnedSpace - m_contentItem->m_leftPinnedSpace,
                             height());
 
-        if (!m_mouseDown) {
+        if (!m_dragging) {
             if (!contentsRect.contains(mappedCurrent)) {
                 m_contentItem->m_viewAnchorItem = m_currentItem;
                 if (qApp->layoutDirection() == Qt::RightToLeft) {
@@ -1087,6 +1149,7 @@ void ColumnView::setCurrentIndex(int index)
                 }
             } else {
                 m_contentItem->snapToItem();
+                m_dragging = false;
             }
         }
     }
@@ -1237,12 +1300,13 @@ void ColumnView::setInteractive(bool interactive)
     m_interactive = interactive;
 
     if (!m_interactive) {
+        m_contentItem->snapToItem();
+
         if (m_dragging) {
             m_dragging = false;
             Q_EMIT draggingChanged();
         }
 
-        m_contentItem->snapToItem();
         setKeepMouseGrab(false);
     }
 
@@ -1262,16 +1326,6 @@ void ColumnView::setAcceptsMouse(bool accepts)
 
     m_acceptsMouse = accepts;
 
-    if (!m_acceptsMouse) {
-        if (m_dragging) {
-            m_dragging = false;
-            Q_EMIT draggingChanged();
-        }
-
-        m_contentItem->snapToItem();
-        setKeepMouseGrab(false);
-    }
-
     Q_EMIT acceptsMouseChanged();
 }
 
@@ -1308,6 +1362,12 @@ void ColumnView::insertItem(int pos, QQuickItem *item)
     connect(attached, &ColumnViewAttached::globalHeaderChanged, m_contentItem, &ContentItem::connectHeader);
     connect(attached, &ColumnViewAttached::globalFooterChanged, m_contentItem, &ContentItem::connectFooter);
 
+    if (attached->interactiveResizeEnabled()) {
+        const qreal preferredWidth = m_state.value(pos);
+        if (preferredWidth > 0) {
+            attached->setPreferredWidth(preferredWidth);
+        }
+    }
     // Animate shift to new item.
     m_contentItem->m_shouldAnimate = true;
     m_contentItem->layoutItems();
@@ -1445,6 +1505,84 @@ QQuickItem *ColumnView::removeItem(QQuickItem *item)
     return item;
 }
 
+QString ColumnView::savedState()
+{
+    int i = 0;
+    QJsonObject obj;
+    for (auto it = m_state.constBegin(); it != m_state.constEnd(); it++) {
+        obj.insert(QString::number(it.key()), it.value());
+    }
+    for (QQuickItem *item : std::as_const(m_contentItem->m_items)) {
+        ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
+        if (!attached->interactiveResizeEnabled()) {
+            ++i;
+            continue;
+        }
+
+        obj.insert(QString::number(i), attached->preferredWidth());
+        m_state[i] = attached->preferredWidth();
+        ++i;
+    }
+
+    QJsonDocument doc(obj);
+
+    return QString::fromUtf8(doc.toJson(QJsonDocument::Compact));
+}
+
+void ColumnView::setSavedState(const QString &state)
+{
+    QJsonParseError parseError;
+    QJsonDocument doc = QJsonDocument::fromJson(state.toUtf8(), &parseError);
+    if (parseError.error != QJsonParseError::NoError) {
+        qCWarning(KirigamiLayoutsLog) << "Error reading state:" << parseError.errorString();
+        return;
+    }
+
+    if (!doc.isObject()) {
+        qCWarning(KirigamiLayoutsLog) << "Error reading state: Not a JSon Object";
+        return;
+    }
+
+    QJsonObject obj = doc.object();
+
+    if (obj.size() == 0) {
+        qCWarning(KirigamiLayoutsLog) << "Error reading state: Empty JSon Object";
+        return;
+    }
+
+    for (auto it = obj.constBegin(); it != obj.constEnd(); it++) {
+        bool ok;
+        int index = it.key().toInt(&ok);
+        if (!ok || index < 0) {
+            qCWarning(KirigamiLayoutsLog) << "Key" << it.key() << "is not a positive integer";
+            continue;
+        }
+        qreal value = it.value().toDouble();
+        if (!it.value().isDouble() || value <= 0) {
+            qCWarning(KirigamiLayoutsLog) << "Value for key" << it.key() << "is not a positive number:" << it.value();
+            continue;
+        }
+
+        m_state[index] = value;
+
+        if (index >= m_contentItem->m_items.count()) {
+            continue;
+        }
+
+        QQuickItem *item = m_contentItem->m_items[index];
+        ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
+        if (!attached->interactiveResizeEnabled()) {
+            continue;
+        }
+
+        attached->setPreferredWidth(value);
+    }
+
+    polish();
+
+    Q_EMIT savedStateChanged();
+}
+
 QQuickItem *ColumnView::removeItem(const int index)
 {
     if (m_contentItem->m_items.isEmpty() || index < 0 || index >= count()) {
@@ -1465,6 +1603,15 @@ QQuickItem *ColumnView::removeItem(const QVariant &item)
     }
 }
 
+QQuickItem *ColumnView::get(int index)
+{
+    if (index < 0 || index > m_contentItem->m_items.count() - 1) {
+        return nullptr;
+    }
+
+    return m_contentItem->m_items.value(index);
+}
+
 QQuickItem *ColumnView::pop(const QVariant &item)
 {
     if (item.canConvert<QQuickItem *>()) {
@@ -1549,13 +1696,14 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
     }
 
     switch (event->type()) {
-    case QEvent::MouseButtonPress: {
-        QMouseEvent *me = static_cast<QMouseEvent *>(event);
+    case QEvent::TouchBegin: {
+        QTouchEvent *te = static_cast<QTouchEvent *>(event);
 
-        if (me->button() != Qt::LeftButton) {
+        if (te->pointCount() != 1) {
             return false;
         }
 
+        m_verticalScrollIntercepted = false;
         // On press, we set the current index of the view to the root item
         QQuickItem *candidateItem = item;
         while (candidateItem->parentItem() && candidateItem->parentItem() != m_contentItem) {
@@ -1565,51 +1713,37 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             setCurrentIndex(idx);
         }
 
-        // if !m_acceptsMouse we don't drag with mouse
-        if (!m_acceptsMouse && me->source() == Qt::MouseEventNotSynthesized) {
-            event->setAccepted(false);
-            return false;
-        }
+        // m_acceptsMouse is a noop now
 
         m_contentItem->m_slideAnim->stop();
         if (item->property("preventStealing").toBool()) {
             m_contentItem->snapToItem();
             return false;
         }
-        m_oldMouseX = m_startMouseX = mapFromItem(item, me->position()).x();
-        m_oldMouseY = m_startMouseY = mapFromItem(item, me->position()).y();
 
-        m_mouseDown = true;
-        me->setAccepted(false);
-        setKeepMouseGrab(false);
+        te->setAccepted(false);
 
         break;
     }
-    case QEvent::MouseMove: {
-        QMouseEvent *me = static_cast<QMouseEvent *>(event);
+    case QEvent::TouchUpdate: {
+        QTouchEvent *te = static_cast<QTouchEvent *>(event);
 
-        if (!m_acceptsMouse && me->source() == Qt::MouseEventNotSynthesized) {
+        if (te->pointCount() != 1) {
             return false;
         }
 
-        if (!(me->buttons() & Qt::LeftButton)) {
+        const QPointF pressPos = mapFromItem(item, te->points().first().pressPosition());
+        const QPointF lastPos = mapFromItem(item, te->points().first().lastPosition());
+        const QPointF pos = mapFromItem(item, te->points().first().position());
+
+        m_verticalScrollIntercepted = m_verticalScrollIntercepted || std::abs(pos.y() - pressPos.y()) > qApp->styleHints()->startDragDistance() * 3;
+
+        if (m_verticalScrollIntercepted) {
+            m_contentItem->snapToItem();
+            m_dragging = false;
             return false;
         }
 
-        // It's possible the synthetyzed mouse press event was not passed
-        // but now we get the mouse move events, consider the first move
-        // as the press
-        if (!m_mouseDown) {
-            m_mouseDown = true;
-            m_oldMouseX = m_startMouseX = mapFromItem(item, me->position()).x();
-            m_oldMouseY = m_startMouseY = mapFromItem(item, me->position()).y();
-            return false;
-        }
-
-        const QPointF pos = mapFromItem(item, me->position());
-
-        bool verticalScrollIntercepted = false;
-
         QQuickItem *candidateItem = item;
         while (candidateItem->parentItem() && candidateItem->parentItem() != m_contentItem) {
             candidateItem = candidateItem->parentItem();
@@ -1625,26 +1759,25 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(candidateItem, true));
 
             ScrollIntentionEvent scrollIntentionEvent;
-            scrollIntentionEvent.delta = QPointF(pos.x() - m_oldMouseX, pos.y() - m_oldMouseY);
+            scrollIntentionEvent.delta = pos - lastPos;
 
             Q_EMIT attached->scrollIntention(&scrollIntentionEvent);
 
             if (scrollIntentionEvent.accepted) {
-                verticalScrollIntercepted = true;
+                m_verticalScrollIntercepted = true;
                 event->setAccepted(true);
             }
         }
 
-        if ((!keepMouseGrab() && (item->keepMouseGrab() || item->keepTouchGrab())) || item->property("preventStealing").toBool()) {
+        if (item->keepTouchGrab() || item->property("preventStealing").toBool()) {
             m_contentItem->snapToItem();
-            m_oldMouseX = pos.x();
-            m_oldMouseY = pos.y();
+            m_dragging = false;
             return false;
         }
 
         const bool wasDragging = m_dragging;
         // If a drag happened, start to steal all events, use startDragDistance * 2 to give time to widgets to take the mouse grab by themselves
-        m_dragging = keepMouseGrab() || qAbs(mapFromItem(item, me->position()).x() - m_startMouseX) > qApp->styleHints()->startDragDistance() * 3;
+        m_dragging = !m_verticalScrollIntercepted && std::abs(pos.x() - pressPos.x()) > qApp->styleHints()->startDragDistance() * 3;
 
         if (m_dragging != wasDragging) {
             m_moving = true;
@@ -1652,25 +1785,42 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             Q_EMIT draggingChanged();
         }
 
+        m_contentItem->m_lastDragDelta = pos.x() - lastPos.x();
+
         if (m_dragging) {
-            m_contentItem->setBoundedX(m_contentItem->x() + pos.x() - m_oldMouseX);
+            m_contentItem->setBoundedX(m_contentItem->x() + m_contentItem->m_lastDragDelta);
         }
 
-        m_contentItem->m_lastDragDelta = pos.x() - m_oldMouseX;
-        m_oldMouseX = pos.x();
-        m_oldMouseY = pos.y();
+        te->setAccepted(m_dragging);
 
-        setKeepMouseGrab(m_dragging);
-        me->setAccepted(m_dragging);
-
-        return m_dragging && !verticalScrollIntercepted;
+        return m_dragging;
     }
-    case QEvent::MouseButtonRelease: {
-        QMouseEvent *me = static_cast<QMouseEvent *>(event);
+    case QEvent::TouchEnd:
+    case QEvent::TouchCancel: {
+        QTouchEvent *te = static_cast<QTouchEvent *>(event);
         if (item->property("preventStealing").toBool()) {
             return false;
         }
 
+        m_verticalScrollIntercepted = false;
+        // if a drag happened, don't pass the event
+        const bool block = m_dragging;
+
+        m_contentItem->snapToItem();
+
+        if (m_dragging) {
+            m_contentItem->m_lastDragDelta = 0;
+            m_dragging = false;
+            Q_EMIT draggingChanged();
+        }
+
+        event->accept();
+
+        te->setAccepted(block);
+        return block;
+    }
+    case QEvent::MouseButtonRelease: {
+        QMouseEvent *me = static_cast<QMouseEvent *>(event);
         if (me->button() == Qt::BackButton && m_currentIndex > 0) {
             setCurrentIndex(m_currentIndex - 1);
             me->accept();
@@ -1680,32 +1830,7 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             me->accept();
             return true;
         }
-
-        if (!m_acceptsMouse && me->source() == Qt::MouseEventNotSynthesized) {
-            return false;
-        }
-
-        if (me->button() != Qt::LeftButton) {
-            return false;
-        }
-
-        m_mouseDown = false;
-
-        if (m_dragging) {
-            m_contentItem->snapToItem();
-            m_contentItem->m_lastDragDelta = 0;
-            m_dragging = false;
-            Q_EMIT draggingChanged();
-        }
-
-        event->accept();
-
-        // if a drag happened, don't pass the event
-        const bool block = keepMouseGrab();
-        setKeepMouseGrab(false);
-
-        me->setAccepted(block);
-        return block;
+        break;
     }
     default:
         break;
@@ -1714,6 +1839,49 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
     return QQuickItem::childMouseEventFilter(item, event);
 }
 
+bool ColumnView::event(QEvent *event)
+{
+    if (!m_interactive) {
+        return QQuickItem::event(event);
+    }
+    switch (event->type()) {
+    case QEvent::TouchBegin: {
+        QTouchEvent *te = static_cast<QTouchEvent *>(event);
+
+        if (te->pointCount() != 1) {
+            m_contentItem->snapToItem();
+            m_dragging = false;
+        }
+        break;
+    }
+    case QEvent::TouchUpdate: {
+        QTouchEvent *te = static_cast<QTouchEvent *>(event);
+
+        if (te->pointCount() != 1) {
+            m_contentItem->snapToItem();
+            m_dragging = false;
+            break;
+        }
+
+        m_dragging = true;
+
+        const QEventPoint point = te->points().first();
+
+        m_contentItem->m_lastDragDelta = point.position().x() - point.lastPosition().x();
+        m_contentItem->setBoundedX(m_contentItem->x() + m_contentItem->m_lastDragDelta);
+        break;
+    }
+    case QEvent::TouchEnd:
+    case QEvent::TouchCancel:
+        m_contentItem->snapToItem();
+        m_dragging = false;
+        break;
+    default:
+        break;
+    }
+    return QQuickItem::event(event);
+}
+
 void ColumnView::mousePressEvent(QMouseEvent *event)
 {
     if (!m_acceptsMouse && event->source() == Qt::MouseEventNotSynthesized) {
@@ -1723,50 +1891,7 @@ void ColumnView::mousePressEvent(QMouseEvent *event)
 
     if (event->button() == Qt::BackButton || event->button() == Qt::ForwardButton) {
         event->accept();
-        return;
     }
-
-    if (!m_interactive) {
-        return;
-    }
-
-    m_contentItem->snapToItem();
-    m_oldMouseX = event->position().x();
-    m_startMouseX = event->position().x();
-    m_mouseDown = true;
-    setKeepMouseGrab(false);
-    event->accept();
-}
-
-void ColumnView::mouseMoveEvent(QMouseEvent *event)
-{
-    if (event->buttons() & Qt::BackButton || event->buttons() & Qt::ForwardButton) {
-        event->accept();
-        return;
-    }
-
-    if (!m_interactive) {
-        return;
-    }
-
-    const bool wasDragging = m_dragging;
-    // Same startDragDistance * 2 as the event filter
-    m_dragging = keepMouseGrab() || qAbs(event->position().x() - m_startMouseX) > qApp->styleHints()->startDragDistance() * 2;
-    if (m_dragging != wasDragging) {
-        m_moving = true;
-        Q_EMIT movingChanged();
-        Q_EMIT draggingChanged();
-    }
-
-    setKeepMouseGrab(m_dragging);
-
-    if (m_dragging) {
-        m_contentItem->setBoundedX(m_contentItem->x() + event->pos().x() - m_oldMouseX);
-    }
-
-    m_contentItem->m_lastDragDelta = event->pos().x() - m_oldMouseX;
-    m_oldMouseX = event->pos().x();
-    event->accept();
 }
 
 void ColumnView::mouseReleaseEvent(QMouseEvent *event)
@@ -1780,40 +1905,6 @@ void ColumnView::mouseReleaseEvent(QMouseEvent *event)
         event->accept();
         return;
     }
-
-    m_mouseDown = false;
-
-    if (!m_interactive) {
-        return;
-    }
-
-    m_contentItem->snapToItem();
-    m_contentItem->m_lastDragDelta = 0;
-
-    if (m_dragging) {
-        m_dragging = false;
-        Q_EMIT draggingChanged();
-    }
-
-    setKeepMouseGrab(false);
-    event->accept();
-}
-
-void ColumnView::mouseUngrabEvent()
-{
-    m_mouseDown = false;
-
-    if (m_contentItem->m_slideAnim->state() != QAbstractAnimation::Running) {
-        m_contentItem->snapToItem();
-    }
-    m_contentItem->m_lastDragDelta = 0;
-
-    if (m_dragging) {
-        m_dragging = false;
-        Q_EMIT draggingChanged();
-    }
-
-    setKeepMouseGrab(false);
 }
 
 void ColumnView::classBegin()
